import numpy as np
import pandas as pd
import igraph
import heapq
from typing import Optional, Union
from scipy.sparse import  csr_matrix


def edge_between_score(
    g: igraph.Graph
) -> list:
    score_list = g.edge_betweenness(directed=True, weights='weight')
    norm = max(score_list)
    score_list = [score_list[i]/norm for i in range(len(score_list))]
    return score_list

def common_neighbor_score(
    g: igraph.Graph
) -> list:
    edges = g.get_edgelist()
    score_list = g.similarity_dice(pairs=edges, loops=True)
    norm = max(score_list)
    score_list = [score_list[i]/norm for i in range(len(score_list))]
    return score_list

def cal_similarity(
    g: igraph.Graph,
    lambda_: float,
) -> csr_matrix:
    between = edge_between_score(g)
    common_neighbor = common_neighbor_score(g)
    score_list = [(lambda_*common_neighbor[i] + (1-lambda_)*between[i]) for i in range(len(between)) ]
    return score_list

def get_edge_index(
    g: igraph.Graph,
    lambda_: float=0.5,
    r: float = 0.5, 
) -> list:
    n = int(g.ecount()*r)
    score = cal_similarity(g, lambda_)
    edges = g.get_edgelist()
    index = [edges[i] for i in (heapq.nsmallest(n, range(len(score)), score.__getitem__))]
    return index

def get_mst_edge_index(
    adj: Union[csr_matrix, np.ndarray],
    mst: Union[csr_matrix, np.ndarray],
    groups: Optional[pd.DataFrame] = None,
) -> list:
    global_index_list = []

    if isinstance(adj, csr_matrix):
        adj = adj.todense()
    if isinstance(mst, csr_matrix):
        mst = mst.todense()

    membership = groups['louvain'].cat.codes.values
    for i in range(len(adj)):
        cluster_i = membership[i]
        for j in range(i, len(adj)):
            cluster_j = membership[j]
            if adj[i,j] == 0 or cluster_i == cluster_j or mst[cluster_i, cluster_j] > 0 or mst[cluster_j, cluster_i] > 0:
                continue
            else:
                global_index_list.append((i,j))        
    return global_index_list

def get_his_edge_index(
    adj: Union[csr_matrix, np.ndarray],
    groups: Optional[pd.DataFrame] = None,
    lambda_: float=0.5,
    r: float = 0.5, 
    mode: Optional[str] = None,  
) -> list:
    global_index_list = []

    if isinstance(adj, csr_matrix):
        adj = adj.todense()

    g = igraph.Graph.Weighted_Adjacency(matrix=adj)
    g.vs['global_index'] = [i for i in range(len(adj))] 
    
    if mode == 'cluster':
        if groups is None:
            raise ValueError('missing key parameter \'groups\'')
        vc = igraph.VertexClustering(
            g, membership=groups['louvain'].cat.codes.values,
        )
        nc = vc.sizes()
        for i in range(len(nc)):
            sg = vc.subgraph(i)
            index = get_edge_index(sg, lambda_=lambda_, r=r)
            index = [(sg.vs['global_index'][index[i][0]], sg.vs['global_index'][index[i][1]]) for i in range(len(index))]
            global_index_list.extend(index)
    else:
        index = get_edge_index(g, lambda_=lambda_, r=r)
        global_index_list = [(g.vs['global_index'][index[i][0]], g.vs['global_index'][index[i][1]]) for i in range(len(index))]

    return global_index_list

def prune_graph(
    adj: Union[csr_matrix, np.ndarray],
    edges: list,
) -> csr_matrix:
    
    if isinstance(adj, csr_matrix):
        adjacency = adj.todense()
    else:
        adjacency = adj.copy()
    for edge in edges:
        # if the edge is the cut edge, ignore it
        if np.sum(adjacency, axis=1)[edge[0]] == adjacency[edge[0], edge[1]] or np.sum(adjacency, axis=1)[edge[1]] == adjacency[edge[1], edge[0]]:
            continue
        # if sum(adjacency[edge[0], ]) == adjacency[edge[0], edge[1]] or sum(sum(adjacency[edge[1], :])) == adjacency[edge[1], edge[0]]:
        #     continue
        adjacency[edge[0], edge[1]] = 0
        adjacency[edge[1], edge[0]] = 0

    return csr_matrix(adjacency)



import os
import sys
import time
import logging
import numpy as np
from . import args
import platform
import anndata as ad
# import torch
# import torch.nn.functional as F
from multiprocessing import Pool
from scipy.sparse import csr_matrix
from typing import Optional
from pathlib import Path
from sklearn.decomposition import PCA
from .ppca import PPCA

if platform.system().lower() == 'windows':
    import umap.umap_ as umap
if platform.system().lower() == 'linux':
    import umap

def neighbor_ave_cpu(A, pad):
	if pad==0:
		return A
	ngene, _ = A.shape
	ll = pad * 2 + 1
	B, C, D, E = [np.zeros((ngene + ll, ngene + ll)) for i in range(4)]
	B[(pad + 1):(pad + ngene + 1), (pad + 1):(pad + ngene + 1)] = A[:]
	F = B.cumsum(axis = 0).cumsum(axis = 1)
	C[ll :, ll:] = F[:-ll, :-ll]
	D[ll:, :] = F[:-ll, :]
	E[:, ll:] = F[:, :-ll]
	return (np.around(F + C - D - E, decimals=8)[ll:, ll:] / float(ll * ll))

def random_walk_cpu(A, rp):
	ngene, _ = A.shape
	A = A - np.diag(np.diag(A))
	A = A + np.diag(np.sum(A, axis=0) == 0)
	P = np.divide(A, np.sum(A, axis = 0))
	Q = np.eye(ngene)
	I = np.eye(ngene)
	for i in range(30):
		Q_new = (1 - rp) * I + rp * np.dot(Q, P)
		delta = np.linalg.norm(Q - Q_new)
		Q = Q_new.copy()
		if delta < 1e-6:
			break
	return Q

def impute_cpu(args):
	cell, c, ngene, pad, rp = args
	D = np.loadtxt(cell + '_chr' + c + '.txt')
	A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape = (ngene, ngene)).toarray()
	A = np.log2(A + A.T + 1)
	A = neighbor_ave_cpu(A, pad)
	if rp==-1:
		Q = A[:]
	else:
		Q = random_walk_cpu(A, rp)
	return [cell, Q.reshape(ngene*ngene)]

def impute_embedding(
    data: np.ndarray,
    chromsize: dict,
    res: int = 1000000,
    imp_method: str = None,
    imp_args = None,
    emb_method = None,
    emb_args = None,
):
    """impute and embed scHi-C data:
    data:np.ndarray, the path to the cell files. 
    chromsize: dict, the size of each chromosome used.
    res: int, bin size
    imp_method: str, the imputation method, it can be "convolution_randomwalk" or None(means no imputation)
    imp_args:   dict, the parameter for imputation if the "imputation_method" is not None, including 
                pad(int) for convolution, 
                rp(float) for random walk,
                prct(float) for binaryzation and ncpus(int) for the multiprocessing.
    emb_method: str, the embedding method,it can be "schicptr", "PCA" or "PPCA". 
                "schicptr" :combing PCA and UMAP i.e. the pipeline proposed by scHiCPTR
                "PCA" : using PCA only to reduce dimension, 
                "PPCA" : replacing the imputation PCA with probabilistic PCA. (imp_method should be set to None)
    emb_args: dict, the parameters for embedding, including:
                pca_ncomponent(int)
                umap_neighbour(int)
                umap_mindist(float)
                umap_ncomponent(int)
                umap_metric(str)
                ppca_sigma(float)
    """
    if imp_method not in args.imp_method_list:
        print("Error: Invalid method for imputation, use \"convolution_randomwalk\" or None")
        sys.exit(-1)

    if emb_method not in args.emb_method_list:
        print("Error: Invalid method for embedding, use one of thses:\"PCA\", \"PPCA\", \"schicptr\"")
        sys.exit(-1)

    if data is None:
        logging.error("Missing key parameter '\data'\!")

    if chromsize is None:
        logging.error("Missing key parameter '\chromsize'\!")

    if imp_method is None:
        logging.warning("No imputation is used") #ignore the imp_args
        pad, rp, prct = 1, -1, -1
    else:
        print("Imputing with", imp_method)
        if imp_args is None:
            imp_args = args.imp_args
            logging.warning("Using default imputaion parameters")
        pad, rp, prct = imp_args["pad"], imp_args["rp"], imp_args["prct"]
    ncpus = imp_args["ncpus"]


    if emb_args is None:
        emb_args = args.emb_args
        logging.warning("Using default embedding parameters")

    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        paras = [[cell, c, ngene, pad, rp] for cell in data]
        p = Pool(ncpus)
        result = p.map(impute_cpu, paras)
        p.close()
        index = {x[0]: j for j, x in enumerate(result)}
        Q_concat = np.array([result[index[x]][1] for x in data])
        if prct > -1:
            thres = np.percentile(Q_concat, 100 - prct, axis=1)
            Q_concat = (Q_concat > thres[:, None])
        end_time = time.time()
        print('chromosome', c, ' load and imputation completed,', 'take', end_time - start_time, 'seconds')
        ndim = int(min(Q_concat.shape) * 0.2) - 1
        if emb_method == "PPCA":
            ## PPCA use PPCA to reduce the dimension firstly instead of PCA.
            ppca_obj = PPCA(q = ndim, sigma = emb_args['ppca_sigma'])
            ppca_obj.fit(Q_concat.T, em = False)
            R_reduce = ppca_obj.transform().T
        else:
            pca = PCA(n_components=ndim)
            R_reduce = pca.fit_transform(Q_concat)
        matrix.append(R_reduce)
        #print(c)
    matrix = np.concatenate(matrix, axis=1)

    print("Embedding with", emb_method)
    pca = PCA(n_components=min(matrix.shape) - 1)
    data = pca.fit_transform(matrix)
    data = data[:, 0:emb_args['pca_ncomponent']]
    if emb_method == "schicptr":
        fit = umap.UMAP(n_neighbors=emb_args['umap_neighbour'], 
                        min_dist=emb_args['umap_mindist'],
                        n_components=emb_args['umap_ncomponent'],
                        metric=emb_args['umap_metric'])
        mapper = fit.fit_transform(data) 
    else:
        mapper = data
    print("Embedding completed")
    return mapper



import os
import numpy as np
import pandas as pd
import scanpy as sc
import collections.abc as cabc
import matplotlib.pyplot as plt
from types import MappingProxyType
from typing import Optional, Union, Mapping, Sequence, Any, List
from pathlib import Path
from anndata import AnnData
from matplotlib.axes import Axes
from matplotlib.colors import is_color_like, Colormap
from . import scHiCPTR

# visualize trajectory or cluster-level graph
def vis_trajectory(
    tra: scHiCPTR.Trajectory,
    save_path: str=None,
    plot_cluster_graph: bool = False,
    plot_trajectory: bool = True,
):
    """\
    Plot the cluster-level graph or trajectory graph.
    This uses ForceAtlas2 or igraph's layout algorithms for most layouts [Csardi06]_.
    The implementation references https://github.com/theislab/scanpy. 

    Parameters
    ----------
    tra
        Class Trajectory.
    save_path
        The dir for saving figures.
    plot_cluster_graph
        IF 'False', do not plot cluster-level graph.
    plot_trajectory
        IF 'True', plot trajectory graph.
    """
    coords = cluster_graph(tra, plot=plot_cluster_graph)
    df = tra._trajectory['pseudotime'].values
    cm = plt.cm.get_cmap('jet')
    fig, ax = plt.subplots(figsize=(4.5, 3))
    ax1 = ax.scatter(coords[:, 0], coords[:, 1], s=8, c=df, alpha=.9, cmap=cm)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.colorbar(ax1)
    fig.tight_layout()
    if save_path == None:
        save_path = os.getcwd()
    fig_path = save_path
    fig.savefig(fig_path, dpi=350)
    # plt.show()
    return

# plot cluster-level graph
def cluster_graph(
    tra: scHiCPTR.Trajectory,
    plot: bool = True,
) -> np.array:
    adata = tra2adata(tra)
    sc.tl.paga(adata, groups='louvain')
    sc.pl.paga(adata, color='louvain', plot=plot)
    sc.tl.draw_graph(adata, init_pos='paga')
    return adata.obsm['X_draw_graph_fa']

# transfer class Trajectory to class AnnData
def tra2adata(tra: scHiCPTR.Trajectory)-> AnnData:
    adata = AnnData(tra._data)
    adata.uns['neighbors'] = {}
    adata.uns['louvain'] = {}
    adata.obsp['distances'] = tra._distacnes_umap
    adata.obsp['connectivities'] = tra._connectivities_umap
    adata.obs = tra._trajectory
    return adata




import logging
import numpy as np
from typing import Optional,Union
import scipy
from scipy.sparse import csr_matrix, issparse

# The code references https://github.com/theislab/scanpy

class DPT():
    """\
    """
    def __init__(self, connectivities):
        self.W = connectivities
        self.Z = None
        self._transitions_sym = None
        self._eigen_values = None
        self._eigen_basis = None
        self._dpt = None

    def compute_transitions(self, density_normalize: bool = True):
        """\
        """
        print("======= computing transitions ========") 
        if density_normalize:
            q = np.asarray(self.W.sum(axis=0))
            if not issparse(self.W):
                Q = np.diag(1.0/q)
            else:
                Q = scipy.sparse.spdiags(1.0/q, 0, self.W.shape[0], self.W.shape[0])
            K = Q @ self.W @ Q
        else:
            K = self.W

        z = np.sqrt(np.asarray(K.sum(axis=0)))
        if not issparse(K):
            self.Z = np.diag(1.0/z)
        else:
            self.Z = scipy.sparse.spdiags(1.0/z, 0, K.shape[0], K.shape[0])
        self._transitions_sym = self.Z @ K @ self.Z
        print("====== finished ======")  
    
    def compute_eigen(
        self,
        n_comps: int = 20,
        sym: Optional[bool] = None,
        sort: str = 'decrease',
    ):
        """\
        """
        np.set_printoptions(precision=10)
        if self._transitions_sym is None:
            raise ValueError('Run `.compute_transitions` first.')
        matrix = self._transitions_sym
        # compute the spectrum
        if n_comps == 0:
            evals, evecs = scipy.linalg.eigh(matrix)
        else:
            n_comps = min(matrix.shape[0]-1, n_comps)
            # ncv = max(2 * n_comps + 1, int(np.sqrt(matrix.shape[0])))
            ncv = None
            which = 'LM' if sort == 'decrease' else 'SM'
            # it pays off to increase the stability with a bit more precision
            matrix = matrix.astype(np.float64)
            evals, evecs = scipy.sparse.linalg.eigsh(
                matrix, k=n_comps, which=which, ncv=ncv
            )
            evals, evecs = evals.astype(np.float32), evecs.astype(np.float32)
        if sort == 'decrease':
            evals = evals[::-1]
            evecs = evecs[:, ::-1]
        logging.info(
            '    eigenvalues of transition matrix\n'
            '    {}'.format(str(evals).replace('\n', '\n    '))
        )
        # if self._number_connected_components > len(evals)/2:
        #     logg.warning('Transition matrix has many disconnected components!')
        self._eigen_values = evals
        self._eigen_basis = evecs

    def get_dpt_row(self, root: int=0):
        mask = None
        eigen_values = self._eigen_values
        eigen_basis = self._eigen_basis
        row = sum(
            (
                eigen_values[l] / (1-eigen_values[l])
                * (eigen_basis[root, l] - eigen_basis[:, l])
            )**2
        # account for float32 precision
            for l in range(0, eigen_values.size)
            if eigen_values[l] < 0.9994
        )
        # thanks to Marius Lange for pointing Alex to this:
        # we will likely remove the contributions from the stationary state below when making
        # backwards compat breaking changes, they originate from an early implementation in 2015
        # they never seem to have deteriorated results, but also other distance measures (see e.g.
        # PAGA paper) don't have it, which makes sense
        row += sum(
            (eigen_basis[root, l] - eigen_basis[:, l])**2
            for l in range(0, eigen_values.size)
            if eigen_values[l] >= 0.9994
        )
        if mask is not None:
            row[~mask] = np.inf
        _dpt = np.sqrt(row)
        self._dpt = _dpt / np.max(_dpt[_dpt < np.inf])

def compute_dpt(
    adj: Union[csr_matrix, None] = None,
    start: int = 0,
):
    dpt = DPT(adj)
    dpt.compute_transitions()
    dpt.compute_eigen()
    dpt.get_dpt_row(root=start)
    return dpt._dpt



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import igraph
from sklearn.metrics import roc_curve, auc
from scipy.stats import kendalltau
from scipy.sparse import csr_matrix
from collections import Counter
from typing import Optional, Union
from sklearn.metrics.cluster import normalized_mutual_info_score as nmi_score
from sklearn.metrics import adjusted_rand_score as ari_score
from sklearn import metrics
from munkres import Munkres, print_matrix

def computing_AUC(rankList, key:str):
    y_true = rankList['bench']
    y_score = np.max(rankList[key])-rankList[key]
    fpr, tpr, threshold = roc_curve(y_true, y_score)
    roc_auc = auc(fpr, tpr)
    if roc_auc < 0.5:
        roc_auc = 1- roc_auc
    # plt.plot(fpr, tpr)
    return roc_auc

def evaluation_ranks_AUC(true_phases:list, true_cell_phases:list, predict_order:pd.DataFrame, key: str) -> pd.DataFrame:
    roc_auc_DF = list()

    
    stage_ind = 'true_cell_phases'
    predict_order[stage_ind] = true_cell_phases

    for k, stage in enumerate(true_phases):
        stage1 = stage
        stage2 = true_phases[(k+1)%len(true_phases)]

        rank_list = predict_order.loc[(predict_order[stage_ind]==stage1)|(predict_order[stage_ind]==stage2)]
        rank_list['bench'] = 0
        rank_list.loc[rank_list[stage_ind]==stage2, 'bench'] = 1
        roc_auc = computing_AUC(rank_list, key)
        # print(stage1, stage2, roc_auc)
        roc_auc_DF.append(roc_auc)
    for k, stage in enumerate(true_phases):
        print(stage,true_phases[(k+1)%len(true_phases)], roc_auc_DF[k])
    roc_auc_DF=pd.DataFrame(roc_auc_DF,index=true_phases)    
    return roc_auc_DF

def compute_tau_b(x, y):
    coef, p = kendalltau(x, y)
    return coef, p

def kendall_rank_coefficient(true_order: list, predict_order: pd.DataFrame, key: str='pseudotime'): 
    true_order_sorted = sorted(true_order)
    # rank = dict(Counter(true_order_sorted))
    predict_order['index'] = np.array(range(len(predict_order)))
    predict_order = predict_order.sort_values(by=[key])
    predict_order['rank'] = np.array(true_order_sorted)
    predict_order = predict_order.sort_values(by='index')
    coef, p = compute_tau_b(true_order, predict_order['rank'].values)
    print(coef, p)
    return coef, p

def modularity_cal(
    adj: Union[csr_matrix, np.ndarray],
    groups: Optional[pd.DataFrame] = None,
) -> float:
    
    if isinstance(adj, csr_matrix):
        adj = adj.todense()
    
    g = igraph.Graph.Weighted_Adjacency(matrix=adj)
    Q = g.modularity(membership=groups['louvain'].cat.codes.values, weights='weight')
    print(Q)
    return Q

def cluster_acc(y_true, y_pred):
    y_true = y_true - np.min(y_true)

    l1 = list(set(y_true))
    numclass1 = len(l1)

    l2 = list(set(y_pred))
    numclass2 = len(l2)

    ind = 0
    if numclass1 != numclass2:
        for i in l1:
            if i in l2:
                pass
            else:
                y_pred[ind] = i
                ind += 1

    l2 = list(set(y_pred))
    numclass2 = len(l2)

    if numclass1 != numclass2:
        print('error')
        return

    cost = np.zeros((numclass1, numclass2), dtype=int)
    for i, c1 in enumerate(l1):
        mps = [i1 for i1, e1 in enumerate(y_true) if e1 == c1]
        for j, c2 in enumerate(l2):
            mps_d = [i1 for i1 in mps if y_pred[i1] == c2]
            cost[i][j] = len(mps_d)

    # match two clustering results by Munkres algorithm
    m = Munkres()
    cost = cost.__neg__().tolist()
    indexes = m.compute(cost)

    # get the match results
    new_predict = np.zeros(len(y_pred))
    for i, c in enumerate(l1):
        # correponding label in l2:
        c2 = l2[indexes[i][1]]

        # ai is the index with label==c2 in the pred_label list
        ai = [ind for ind, elm in enumerate(y_pred) if elm == c2]
        new_predict[ai] = c

    acc = metrics.accuracy_score(y_true, new_predict)
    f1_macro = metrics.f1_score(y_true, new_predict, average='macro')
    precision_macro = metrics.precision_score(y_true, new_predict, average='macro')
    recall_macro = metrics.recall_score(y_true, new_predict, average='macro')
    f1_micro = metrics.f1_score(y_true, new_predict, average='micro')
    precision_micro = metrics.precision_score(y_true, new_predict, average='micro')
    recall_micro = metrics.recall_score(y_true, new_predict, average='micro')
    return acc, f1_macro

def cluster_eva(y_true, y_pred, epoch=0):
    # accuracy, F1
    acc, f1 = cluster_acc(y_true, y_pred)
    # NMI
    nmi = nmi_score(y_true, y_pred, average_method='arithmetic')
    # ARI
    ari = ari_score(y_true, y_pred)
    print(epoch, ':acc {:.4f}'.format(acc), ', nmi {:.4f}'.format(nmi), ', ari {:.4f}'.format(ari),
            ', f1 {:.4f}'.format(f1))



import numpy as np
from numpy.linalg import inv
from numpy import transpose as tr
#import ipdb


class PPCA(object):
    def __init__(self, q=2, sigma=1.0):
        self.q = q
        self.prior_sigma = sigma

    def fit(self, y, em=False):
        self.y = y
        self.p = y.shape[0]
        self.n = y.shape[1]
        if em:
            [self.w, self.mu, self.sigma] = self.__fit_em()
        else:
            [self.w, self.mu, self.sigma] = self.__fit_ml()

    def transform(self, y=None):
        if y is None:
            y = self.y
        [w, mu, sigma] = [self.w, self.mu, self.sigma]
        m = tr(w).dot(w) + sigma * np.eye(w.shape[1])
        m = inv(m)
        x = m.dot(tr(w)).dot(y - mu)
        return x

    def fit_transform(self, *args, **kwargs):
        self.fit(*args, **kwargs)
        return self.transform()

    def transform_infers(self, x=None, noise=False):
        if x is None:
            x = self.transform()
        [w, mu, sigma] = [self.w, self.mu, self.sigma]
        y = w.dot(x) + mu
        if noise:
            for i in range(y.shape[1]):
                e = np.random.normal(0, sigma, y.shape[0])
                y[:, i] += e
        return y

    def __ell(self, w, mu, sigma, norm=True):
        m = inv(tr(w).dot(w) + sigma * np.eye(w.shape[1]))
        mw = m.dot(tr(w))
        ll = 0.0
        for i in range(self.n):
            yi = self.y[:, i][:, np.newaxis]
            yyi = yi - mu
            xi = mw.dot(yyi)
            xxi = sigma * m + xi.dot(tr(xi))
            ll += 0.5 * np.trace(xxi)
            if sigma > 1e-5:
                ll += (2 * sigma)**-1 * float(tr(yyi).dot(yyi))
                ll -= sigma**-1 * float(tr(xi).dot(tr(w)).dot(yyi))
                ll += (2 * sigma)**-1 * np.trace(tr(w).dot(w).dot(xxi))
        if sigma > 1e-5:
            ll += 0.5 * self.n * self.p * np.log(sigma)
        ll *= -1.0
        if norm:
            ll /= float(self.n)
        return ll

    def __fit_em(self, maxit=20):
        w = np.random.rand(self.p, self.q)
        mu = np.mean(self.y, 1)[:, np.newaxis]
        sigma = self.prior_sigma
        ll = self.__ell(w, mu, sigma)

        yy = self.y - mu
        s = self.n**-1 * yy.dot(tr(yy))

        ## 
        del yy
        self.y = 0
        
        for i in range(maxit):
            m = inv(tr(w).dot(w) + sigma * np.eye(self.q))
            t = inv(sigma * np.eye(self.q) + m.dot(tr(w)).dot(s).dot(w))
            w_new = s.dot(w).dot(t)
            sigma_new = self.p**-1 * np.trace(s - s.dot(w).dot(m).dot(tr(w_new)))
            ll_new = self.__ell(w_new, mu, sigma_new)
            print ("{:3d}  {:.3f}".format(i + 1, ll_new))
            w = w_new
            sigma = sigma_new
            ll = ll_new
        return (w, mu, sigma)

    def __fit_ml(self):
        mu = np.mean(self.y, 1)[:, np.newaxis]
        [u, s, v] = np.linalg.svd(self.y - mu)
        if self.q > len(s):
            ss = np.zeros(self.q)
            ss[:len(s)] = s
        else:
            ss = s[:self.q]
        ss = np.sqrt(np.maximum(0, ss**2 - self.prior_sigma))
        w = u[:, :self.q].dot(np.diag(ss))
        if self.q < self.p:
            sigma = 1.0 / (self.p - self.q) * np.sum(s[self.q:]**2)
        else:
            sigma = 0.0
        return (w, mu, sigma)


from numba.cuda import args
import numpy as np
import pandas as pd
import louvain
import igraph
import logging
from . import args, embs, prune, dpt
import scipy as sp
from typing import List, Optional, NamedTuple, Union
from natsort import natsorted
from scipy.sparse import spmatrix, csr_matrix, coo_matrix
from sklearn.metrics import pairwise_distances
from scipy.sparse.csgraph import minimum_spanning_tree
import warnings
warnings.filterwarnings("ignore")


def get_indices_distances_from_dense_matrix(D, n_neighbors: int):
    sample_range = np.arange(D.shape[0])[:, None]
    indices = np.argpartition(D, n_neighbors-1, axis=1)[:, :n_neighbors]
    indices = indices[sample_range, np.argsort(D[sample_range, indices])]
    distances = D[sample_range, indices]
    return indices, distances

def get_sparse_matrix_from_indices_distances_numpy(indices, distances, n_obs, n_neighbors):
    n_nonzero = n_obs * n_neighbors
    indptr = np.arange(0, n_nonzero + 1, n_neighbors)
    D = csr_matrix((
        distances.copy().ravel(),  # copy the data, otherwise strange behavior here
        indices.copy().ravel(),
        indptr,
    ), shape=(n_obs, n_obs))
    D.eliminate_zeros()
    return D

def get_sparse_matrix_from_indices_distances_umap(knn_indices, knn_dists, n_obs, n_neighbors):
    rows = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    cols = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    vals = np.zeros((n_obs * n_neighbors), dtype=np.float64)

    for i in range(knn_indices.shape[0]):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            if knn_indices[i, j] == i:
                val = 0.0
            else:
                val = knn_dists[i, j]

            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val

    result = coo_matrix((vals, (rows, cols)),
                                      shape=(n_obs, n_obs))
    result.eliminate_zeros()
    return result.tocsr()

def compute_connectivities_umap(
    knn_indices, knn_dists,
    n_obs, n_neighbors, set_op_mix_ratio=1.0,
    local_connectivity=1.0,
):
    """\
    This is from umap.fuzzy_simplicial_set [McInnes18]_.

    Given a set of data X, a neighborhood size, and a measure of distance
    compute the fuzzy simplicial set (here represented as a fuzzy graph in
    the form of a sparse matrix) associated to the data. This is done by
    locally approximating geodesic distance at each point, creating a fuzzy
    simplicial set for each such point, and then combining all the local
    fuzzy simplicial sets into a global one via a fuzzy union.
    """
    from umap.umap_ import fuzzy_simplicial_set

    X = coo_matrix(([], ([], [])), shape=(n_obs, 1))
    connectivities = fuzzy_simplicial_set(
        X,
        n_neighbors,
        None,
        None,
        knn_indices=knn_indices,
        knn_dists=knn_dists,
        set_op_mix_ratio=set_op_mix_ratio,
        local_connectivity=local_connectivity,
    )

    if isinstance(connectivities, tuple):
        # In umap-learn 0.4, this returns (result, sigmas, rhos)
        connectivities = connectivities[0]

    distances = get_sparse_matrix_from_indices_distances_umap(
        knn_indices, knn_dists, n_obs, n_neighbors
    )

    return distances, connectivities.tocsr()

def get_igraph_from_adjacency(adjacency, directed=None):
    """Get igraph graph from adjacency matrix."""
    import igraph as ig

    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shape[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es['weight'] = weights
    except:
        pass
    # if g.vcount() != adjacency.shape[0]:
    #     logg.warning(
    #         f'The constructed graph has only {g.vcount()} nodes. '
    #         'Your adjacency matrix contained redundant nodes.'
    #     )
    return g

def get_connectivities_tree(connectivities):
    inverse_connectivities = connectivities.copy()
    inverse_connectivities.data = 1./inverse_connectivities.data
    connectivities_tree = minimum_spanning_tree(inverse_connectivities)
    connectivities_tree_indices = [
        connectivities_tree[i].nonzero()[1]
        for i in range(connectivities_tree.shape[0])]
    connectivities_tree = sp.sparse.lil_matrix(connectivities.shape, dtype=float)
    for i, neighbors in enumerate(connectivities_tree_indices):
        if len(neighbors) > 0:
            connectivities_tree[i, neighbors] = connectivities[i, neighbors]
    return connectivities_tree.tocsr()

def get_sparse_from_igraph(graph, weight_attr=None):
    from scipy.sparse import csr_matrix

    edges = graph.get_edgelist()
    if weight_attr is None:
        weights = [1] * len(edges)
    else:
        weights = graph.es[weight_attr]
    if not graph.is_directed():
        edges.extend([(v, u) for u, v in edges])
        weights.extend(weights)
    shape = graph.vcount()
    shape = (shape, shape)
    if len(edges) > 0:
        return csr_matrix((weights, zip(*edges)), shape=shape)
    else:
        return csr_matrix(shape)

class Trajectory:
    """\
    """
    def __init__(self, embedding: np.ndarray):
        self._data: np.ndarray = embedding
        self._distacnes_knn: Union[np.ndarray, csr_matrix, None] = None
        self._distacnes_umap: Union[np.ndarray, csr_matrix, None] = None
        self._trajectory: pd.DataFrame = pd.DataFrame()
        self._connectivities_umap: Union[np.ndarray, csr_matrix, None] = None
        self._connectivities_prune: Union[np.ndarray, csr_matrix, None] = None
        self._connectivities_cluster: Union[np.ndarray, csr_matrix, None] = None
        self._inter_es: Union[np.ndarray, csr_matrix, None] = None
        self._connectivities_tree: Union[np.ndarray, csr_matrix, None] = None
        self._HIS_index: Union[list, None] = None
        self._MST_inex: Union[list, None] = None
    
    def _compute_graph(
        self,
        n_neighbors: int = 15,
        metric: Union[str, None] = 'euclidean'  
    ) -> None:
        """\
        """
        X = self._data
        if n_neighbors > X.shape[0]:
            n_neighbors = 1 + int(0.5*X.shape[0])
        self.n_neighbors = n_neighbors
        # neighbor search
        # use_dense_distances = (metric == 'euclidean' and self._data.shape[0] < 8912) 
        # if use_dense_distances:
        distances = pairwise_distances(X, metric=metric)
        knn_indices, knn_distances = get_indices_distances_from_dense_matrix(distances, n_neighbors)
        self._distacnes_knn = get_sparse_matrix_from_indices_distances_numpy(knn_indices, knn_distances, X.shape[0], n_neighbors)
        self._distacnes_umap, self._connectivities_umap = compute_connectivities_umap(knn_indices, knn_distances, X.shape[0], n_neighbors)
        return

    def _detect_communities(
        self,
        resolution: float = 0.5,
        random_state: Union[None, int] = 0,
        use_weights: bool = False,
        partition_type = None,
        **Kwargs,
    ) -> None:
        if self._connectivities_umap is None:
            logging.error("No KNN graph!")
        adjacency = self._connectivities_umap
        g = get_igraph_from_adjacency(adjacency, directed = True)
        if use_weights:
            weights = np.array(g.es["weight"]).astype(np.float64)
        else:
            weights = None
        # louvain
        if partition_type == None:
            partition_type = louvain.RBConfigurationVertexPartition
        if resolution is not None:
            Kwargs['resolution_parameter'] = resolution
        if use_weights:
            Kwargs['weights'] = weights
        # if version.parse(louvain.__version__) < version.parse('0.7.0'):
        #     louvain.set_rng_seed(random_state)
        part = louvain.find_partition(
            g, partition_type,
            **Kwargs,
        )
        groups = np.array(part.membership)
        self.louvain_resolution = resolution
        self.louvain_random_state = random_state
        self._trajectory['louvain'] = pd.Categorical(
            values = groups.astype('U'),
            categories = natsorted(map(str, np.unique(groups)))
        )
    
    # MST on PAGA-connectivities-based cluster graph
    def _abstract_cluster_tree(
        self,
    ) -> None:
        ones = self._distacnes_umap
        ones.data = np.ones(len(ones.data))
        g = get_igraph_from_adjacency(ones, directed=True)
        vc = igraph.VertexClustering(
            g,
            membership=self._trajectory['louvain'].cat.codes.values,
        )
        ns = vc.sizes()
        n = sum(ns)
        es_inner_cluster = [vc.subgraph(i).ecount() for i in range(len(ns))]
        cg = vc.cluster_graph(combine_edges='sum')
        inter_es = get_sparse_from_igraph(cg, weight_attr='weight')
        es = np.array(es_inner_cluster) + inter_es.sum(axis=1).A1
        inter_es = inter_es + inter_es.T  # \epsilon_i + \epsilon_j
        connectivities = inter_es.copy()
        expected_n_edges = inter_es.copy()
        inter_es = inter_es.tocoo()
        for i, j, v in zip(inter_es.row, inter_es.col, inter_es.data):
            expected_random_null = (es[i]*ns[j] + es[j]*ns[i])/(n - 1)
            if expected_random_null != 0:
                scaled_value = v / expected_random_null
            else:
                scaled_value = 1
            if scaled_value > 1:
                scaled_value = 1
            connectivities[i, j] = scaled_value
            expected_n_edges[i, j] = expected_random_null 

        self.ns = ns
        self._inter_es = inter_es.tocsr
        self._connectivities_cluster = connectivities
        self._connectivities_tree = get_connectivities_tree(connectivities)   

    def _prune_graph(
        self,
        lambda_: float = 0.5,
        threshold: float = 0.5,
        louvain_res: float = 0.5,
        **Kwargs,
    ):
        self._prune_HIS(lambda_=lambda_, threshold=1-threshold)
        self._prune_MST(res=louvain_res, **Kwargs)
        his_index = self._HIS_index
        mst_index = self._MST_inex
        edges_index = list(set(his_index+mst_index))
        self._connectivities_prune = prune.prune_graph(self._connectivities_umap, edges=edges_index)
        return
    
    def _prune_HIS(
        self,
        lambda_: float = 0.5,
        threshold: float = 0.5,
    ) :
        if self._connectivities_umap is None:
            logging.error("No KNN graph!")
        adjacency = self._connectivities_umap
        self._HIS_index = prune.get_his_edge_index(adj=adjacency, lambda_=lambda_, r=threshold)   
        return 

    def _prune_MST(
        self,
        res: float = 0.5,
        **Kwargs,
    ) -> list:
        if self._connectivities_umap is None:
            logging.error("No KNN graph!")
        self._detect_communities(resolution=res, **Kwargs)
        self._abstract_cluster_tree()
        self._MST_inex = prune.get_mst_edge_index(self._connectivities_umap, self._connectivities_tree, self._trajectory)
        return

def scHiCPTR(
    data: np.ndarray,
    imp_method = "convolution_randomwalk",
    chromsize: dict = None,
    resolution: int = 1000000,
    imp_args = None,
    emb_method = "schicptr",
    emb_args = None,
    embedding: np.ndarray = None,
    start: int = 0,
    knn_args: dict = None,
    prune_args: dict = None,
) -> Trajectory:
    
    """Pipeline of imputation, embedding and pseudotime inference :
    data:       np.ndarray, the path to the cell files. 
    imp_method: str, the imputation method, it can be "convolution_randomwalk" or None(means no imputation)
    chromsize:  dict, the size of each chromosome used.
    resolution: int, bin size
    imp_args:   dict, the parameter for imputation if the "imputation_method" is not None, including 
                pad (int) for convolution, 
                rp (float) for random walk,
                prct(float) for binaryzation,
                ncpus(int) for the multiprocessing.
                If it is set to None by default, they will be 3,0.5,20 and 10, respectively.
    emb_method: str, the embedding method,it can be "schicptr", "PCA" or "PPCA". 
                "schicptr" :combing PCA and UMAP i.e. the pipeline proposed by scHiCPTR
                "PCA" : using PCA only to reduce dimension, 
                "PPCA" : replacing the imputation PCA with probabilistic PCA. (imp_method should be set to None)
    emb_args:   dict, the parameters for embedding, including
                pca_ncomponent (int),
                umap_neighbour (int),
                umap_mindist (float),
                umap_ncomponent (int),
                umap_metric (str),
                ppca_sigma (float)
                If it is set to None by default, they will be 120, 50, 0.5, 20, 'euclidean' and 50, respectively.
    embedding:  np.ndarry, with 2 dimensions, the 1st shape is the number of cells and the 2nd shape is the
                dimension of embeddings. the "data" will be ignored if this parameter is not None
    start:      int, the root cell for trajectory inference
    knn_args:   dict, the parameters for knn graph construction, including
                n_neighbors (int)
                metric (str)
                If it is set to None, they will be 100 and 'euclidean', respectively.
    prune_args: dict, the parameters for graph pruning, including
                lambda_ (float)
                threshold (float)
                louvain_res (float)
                If it is set to None, they will be 0.5, 0.7 and 0.5, respectively.
    """

    if data is None and embedding is None:
        logging.error("Missing key parameter '\embedding'\ ")
    
    if (chromsize is None) and (embedding is None):
        logging.warning("Using default chromsize parameters!")
        chromsize = args.chromsize
    
    if knn_args is None:
        logging.warning("Using default knn parameters!")
        knn_args = args.knn_args

    if prune_args is None:
        logging.warning("Using default pruning parameters!")
        prune_args = args.prune_args
    
    if embedding is None:
        embedding = embs.impute_embedding(data, chromsize=chromsize, res=resolution,imp_method=imp_method,
                                        imp_args=imp_args, emb_method=emb_method, emb_args=emb_args)
    tra = Trajectory(embedding)
    tra._compute_graph(**knn_args)
    tra._prune_graph(**prune_args)
    tra._trajectory['pseudotime'] = dpt.compute_dpt(tra._connectivities_prune, start=start)
    return tra



import sys

if sys.version_info < (3, 6):
    sys.exit('scHiCPTR requires Python >= 3.6')
from pathlib import Path
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scHiCPTR",
    version="0.1.2",
    author="",
    author_email="",
    description="An unsupervised pseudotime inference pipeline through dual graph refinement for single cell Hi-C data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lhqxinhun/scHiCPTR",
    project_urls={
        "Bug Tracker": "https://github.com/lhqxinhun/scHiCPTR",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        l.strip() for l in Path('requirements.txt').read_text('utf-8').splitlines()
    ]
)



import numpy as np
from scHiCPTR import scHiCPTR, eva, vis



####### example code for taking contact matrices as input #######
if __name__ == '__main__':

##For the given small dataset "scHiCPTR/data/matrices/1M.zip", please unzip it before use
    ## 1. Prepare dataset
    rawpath = "../data/matrices/1M"
    network = np.array([rawpath + "/cell_" + str(i) for i in list(range(140, 165))+list(range(440, 465))+list(range(700, 725))+list(range(1020, 1045))], dtype=np.str)

    mm9dim = [197195432,181748087,159599783,155630120,152537259,149517037,152524553,131738871,124076172,129993255,121843856,121257530,120284312,125194864,103494974,98319150,95272651,90772031,61342430,166650296]
    chrom =  [str(i+1) for i in range(19)] + ['X']
    chromsize = {chrom[i]:mm9dim[i] for i in range(len(chrom))}

    ## 2. Imputation and embedding args for the small dataset
    imp_args = None # e.g. default
    emb_args = {"pca_ncomponent":50, "umap_neighbour":15, "umap_mindist":0.5, "umap_ncomponent":5, "umap_metric":"euclidean"}
    ## 3. KNN args
    knn_args = {"n_neighbors": 20, "metric":'euclidean'}
    ## 4. Pruning args
    prune_args = {"lambda_": 0.5, "threshold": 0.7, "louvain_res": 0.5}

    ## 5. Calculate pseudotime
    trajectory = scHiCPTR.scHiCPTR(data=network, chromsize=chromsize, resolution=1000000, emb_method="schicptr", 
                                    emb_args=emb_args, embedding=None, knn_args=knn_args, prune_args=prune_args, start=0)
    result = trajectory._trajectory 
    result.to_csv('../data/results/pseudotime_test.csv')




###### example code for taking embedding vectors as input #######
# if __name__ == '__main__':
    
#     ## 1. Load embedding vector
#     embeds = np.loadtxt('../data/embeds/cellcycle_250k.txt' , dtype=np.str)
    
#     ## 2. KNN args
#     knn_args = {"n_neighbors": 100, "metric":'euclidean'}

#     ## 3. Pruning args
#     prune_args = {"lambda_": 0.5, "threshold": 0.7, "louvain_res": 0.5}

#     ## 4. Calculate pseudotime.
#     trajectory = scHiCPTR.scHiCPTR(data=None, embedding=embeds, knn_args=knn_args, prune_args=prune_args, start=270)
#     result = trajectory._trajectory 
#     result.to_csv('../data/results/pseudotime_cellcycle.csv')

#     ## EValuate results with AUC
#     true_phases = ['G1', 'ES', 'MS', 'G2']    # the true order of the pahses, i.e.  between which successive phases, the AUC will be calculated
#     true_cell_phases =  ['G1']*280+['ES']*303+['MS']*262+['G2']*326 #the ground truth phases for each cell
#     eva.evaluation_ranks_AUC(true_phases=true_phases, true_cell_phases=true_cell_phases, predict_order=result, key='pseudotime')

#      ## EValuate results with Kendall rank coefficient
#     true_order = [0]*280+[1]*303+[2]*262+[3]*326 #the true order for the cells, cells in the same phase are regarded as the same order
#     eva.kendall_rank_coefficient(true_order=true_order,predict_order=result, key='pseudotime')

#     ## Save the visualized pseudotime as pseudotime.jpg to the given path
#     vis.vis_trajectory(trajectory, save_path='../data/results/pseudotime_cellcycle.jpg')
